package ch.unibe.piccola.compiler;

import ch.unibe.piccola.code.Code;
import ch.unibe.piccola.code.CodeFactory;
import ch.unibe.piccola.syntax.*;
import ch.unibe.piccola.source.SourceRef;

/**
 * @author   Stefan Kneubuehl
 * @version  3.4c (Nov 14, 2002)
**/
public class CodeMaker extends SimplifiedNodeVisitor
{

    // ------------------------------------------------------------------------
    // public static
    // ------------------------------------------------------------------------

    public static Code apply(Node node)
    {
        CodeMaker cm = new CodeMaker();
        node.accept(cm);
        cm.m_code.emitExit(node.getSourceRef());
        return cm.m_code.code();
    }

    // ------------------------------------------------------------------------
    // private static
    // ------------------------------------------------------------------------

    private static final String LABEL_ARG = Syntax.labelArg();

    private static final String LABEL_DYNAMIC = Syntax.labelDynamic();

    private static final String LABEL_SET = Syntax.labelSet();

    // ------------------------------------------------------------------------
    // public
    // ------------------------------------------------------------------------

    /**
     * \x: body   ==>  \: root = (root, x = dynamic.arg), body
    **/
    public Object visitAbstraction(Abstraction node)
    {
        SourceRef sr = node.getSourceRef();
        CodeMaker cm = new CodeMaker();
        if (node.getParameter().equals("")) {
            node.getChild().accept(cm);
        }
        else {
            cm.m_code.emitRoot(sr);
            cm.m_code.emitLookup(sr, LABEL_DYNAMIC);
            cm.m_code.emitProject(sr, LABEL_ARG);
            cm.m_code.emitBind(sr, node.getParameter());
            cm.m_code.emitExtend(sr);
            cm.m_code.emitPushContext(sr);
            node.getChild().accept(cm);
            cm.m_code.emitPopContext(sr);
        }
        cm.m_code.emitExit(sr);
        m_code.emitService(sr, cm.m_code.code());
        return null;
    }

    public Object visitApplication(Application node)
    {
        SourceRef sr = node.getSourceRef();
        visit(node.getLeftChild());
        m_code.emitLookup(sr, LABEL_DYNAMIC);
        visit(node.getRightChild());
        m_code.emitBind(sr, LABEL_ARG);
        m_code.emitExtend(sr);
        m_code.emitApply(node.getSourceRef());
        return null;
    }

    /**
     * c[bind(x, F)]  ==>  c[F], bind(x)
    **/
    public Object visitBinding(Binding node)
    {
        visit(node.getChild());
        m_code.emitBind(node.getSourceRef(), node.getLabel());
        return null;
    }

    public Object visitConstant(Constant node)
    {
        m_code.emitJavaObject(node.getSourceRef(), node.getConstant());
        return null;
    }

    /**
     * c[empty]  ==>  empty
    **/
    public Object visitEmpty(Empty node)
    {
        m_code.emitEmpty(node.getSourceRef());
        return null;
    }

    /**
     * c[extend(F, G)]  ==>  c[F], c[G], extend
    **/
    public Object visitExtension(Extension node)
    {
        if (node.getLeftChild() instanceof Empty) {
            visit(node.getRightChild());
            return null;
        }
        if (node.getRightChild() instanceof Empty) {
            visit(node.getLeftChild());
            return null;
        }
        visit(node.getLeftChild());
        visit(node.getRightChild());
        m_code.emitExtend(node.getSourceRef());
        return null;
    }

    /**
     * c[fixedPoint(L, F)] =
    **/
    public Object visitFixedPoint(FixedPoint node)
    {
        SourceRef sr = node.getSourceRef();
        String label = node.getLabel();
        // root = (root, I = <VAR>)
        m_code.emitRoot(sr);
        m_code.emitEmpty(sr);
        m_code.emitVariable(sr);
        m_code.emitBind(sr, label);
        m_code.emitExtend(sr);
        m_code.emitPushContext(sr);

        // I.set(I = F).I
        m_code.emitLookup(sr, label);
        m_code.emitProject(sr, LABEL_SET);

        m_code.emitLookup(sr, LABEL_DYNAMIC);
        visit(LookupTransformation.apply(node.getLabel(), node.getChild()));

        m_code.emitBind(sr, label);
        m_code.emitBind(sr, LABEL_ARG);
        m_code.emitExtend(sr);
        m_code.emitApply(sr);
        m_code.emitProject(sr, label);

        // x = (root = (root, I = <VAR>()), (I.set(I = lookup[F]).I)
        m_code.emitPopContext(sr);
        m_code.emitBind(sr, label);
        return null;
    }

    /**
     * c[hide(F, x)]  ==>  c[F], hide(x)
    **/
    public Object visitHide(Hide node)
    {
        visit(node.getChild());
        m_code.emitHide(node.getSourceRef(), node.getLabel());
        return null;
    }

    /**
     * c[label(x)]  ==>  label(x)
    **/
    public Object visitLabel(Label node)
    {
        m_code.emitLookup(node.getSourceRef(), node.getLabel());
        return null;
    }

    public Object visitProjection(Projection node)
    {
        visit(node.getChild());
        m_code.emitProject(node.getSourceRef(), node.getLabel());
        return null;
    }

    public Object visitRoot(Root node)
    {
        m_code.emitRoot(node.getSourceRef());
        return null;
    }

    public Object visitSandbox(Sandbox node)
    {
        SourceRef sr = node.getSourceRef();
        visit(node.getLeftChild());
        if (node.getRightChild() instanceof Empty) {
            m_code.emitPopValue(sr);
            m_code.emitEmpty(sr);
        }
        else {
            m_code.emitPushContext(sr);
            visit(node.getRightChild());
            m_code.emitPopContext(sr);
        }
        return null;
    }

    // ------------------------------------------------------------------------
    // private
    // ------------------------------------------------------------------------

    private CodeFactory m_code;

    private CodeMaker()
    {
        m_code = new CodeFactory();
    }

}

