package ch.unibe.piccola.syntax;

import ch.unibe.piccola.source.SourceRef;

/**
 * Represents a syntax node with two child nodes. The following node types
 * have two children:
 * <ul>
 *   <li>Application
 *   <li>CompoundBinding
 *   <li>Extension
 *   <li>InfixApplication
 *   <li>NamedAbstraction
 *   <li>Quote
 *   <li>Sandbox
 * </ul>
 *
 * @author   Stefan Kneubuehl
 * @version  3.4c (Nov 14, 2002)
**/
public abstract class TwoChildrenNode extends Node
{

    // ------------------------------------------------------------------------
    // public
    // ------------------------------------------------------------------------

    /**
     * Compares this node with the specified object.
     *
     * @param  object the object to be compared with this node.
     *
     * @return <code>true</code> if this node is equal to object,
     *         <code>false</code> otherwise.
    **/
    public boolean equals(Object object)
    {
        if (!super.equals(object)) {
            return false;
        }
        TwoChildrenNode other = (TwoChildrenNode) object;
        return m_leftChild.equals(other.m_leftChild) &&
               m_rightChild.equals(other.m_rightChild);
    }

    /**
     * Returns the left child of this this node.
     *
     * @return the left child of this this node.
    **/
    public Node getLeftChild()
    {
        return m_leftChild;
    }

    /**
     * Returns the right child of this this node.
     *
     * @return the right child of this this node.
    **/
    public Node getRightChild()
    {
        return m_rightChild;
    }

    /**
     * Returns a debug string representation of this node.
     *
     * @return a debug string representation of this node.
    **/
    public String toString()
    {
        return "(" + m_leftChild + ", " + m_rightChild + ")";
    }

    public Node transform(NodeVisitor visitor)
    {
        Node leftChild = (Node) m_leftChild.accept(visitor);
        Node rightChild = (Node) m_rightChild.accept(visitor);
        return (leftChild == m_leftChild && rightChild == m_rightChild) ?
            this : newInstance(leftChild, rightChild);
    }

    // ------------------------------------------------------------------------
    // protected
    // ------------------------------------------------------------------------

    protected TwoChildrenNode(SourceRef sourceRef, Node leftChild,
                              Node rightChild)
    {
        super(sourceRef);
        m_leftChild = leftChild;
        m_rightChild = rightChild;
    }

    protected abstract Node newInstance(Node leftChild, Node rightChild);

    // ------------------------------------------------------------------------
    // private
    // ------------------------------------------------------------------------

    private final Node m_leftChild;

    private final Node m_rightChild;

}