package ch.unibe.piccola.vm;

import java.util.Iterator;
import java.util.SortedMap;
import java.util.TreeMap;
import ch.unibe.piccola.code.Arguments;
import ch.unibe.piccola.syntax.Syntax;
import ch.unibe.piccola.type.ContractualType;
import ch.unibe.piccola.code.Code;

/**
 *
 * @author   Stefan Kneubuehl
 * @version  3.6c (Apr 23, 2003)
**/
public abstract class Form {

    /**
     * For compatibility with ByteCode stuff only.
    **/
    public final Object callback(Arguments args)
    {
        return service().callback(args);
    }

    /**
     * For compatibility with Piccola only.
    **/
    public final void run(Form dynamic)
    {
        service().run(dynamic);
    }

    // ------------------------------------------------------------------------
    // public
    // ------------------------------------------------------------------------

    /**
     * Binds this form to the specified label.
     *
     * @param  label the label this form will be bound to.
     *
     * @return this form bound to the specified label, i.e.
     *         <i>(label = this)</i>.
    **/
    public abstract Form bind(String label);

    /**
     * Downs this form.
     *
     * @return the object representing the downed form.
    **/
    public abstract Object down();

    /**
     * Compares this form to the specified object.
     *
     * @param  object the object to be compared with this form.
     *
     * @return <code>true</code> if this form is equal to object,
     *         <code>false</code> otherwise.
    **/
    public abstract boolean equals(Object object);

    /**
     * Checks whether the specified label is bound by this form.
     *
     * @param  label the label for which the existence will be checked.
     *
     * @return <code>true</code> if this form binds the label,
     *         <code>false</code> otherwise.
    **/
    public abstract boolean exists(String label);

    /**
     * Extends this form with the specified form.
    **/
    public abstract Form extend(Form form);

    /**
     * Returns a sorted map containing all bindings of this form. The keys of
     * the map are the labels represented as String object, the values are the
     * corresponding Forms.
     *
     * @return a SortedMap mapping the labels of this form to their values.
    **/
    public abstract SortedMap getBindings();

    /**
     * Unwraps this form.
     *
     * @return the Java object represented by this form.
    **/
    public abstract Object unwrap();

    /**
     * Returns the hash code of this form.
     *
     * @return the hash code of this form.
    **/
    public abstract int hashCode();

    /**
     * Hides the specified label in this form.
    **/
    public abstract Form hide(String label);

    public abstract Iterator labels();

    public abstract boolean isEmpty();

    /**
     * Checks whether this form is a peer form.
     *
     * @return <code>true</code> if this form is a peer form,
     *         <code>false</code> otherwise.
    **/
    public abstract boolean isPeer();

    /**
     * Checks whether this form is executable (a service).
     *
     * @return <code>true</code> if this form can be executed,
     *         <code>false</code> otherwise.
    **/
    public abstract boolean isService();

    /**
     * Projects this form on the specified label.
     *
     * @param  label the label to be projected.
     *
     * @return The form bound to label in this form, i.e. returns
     *         <i>this.label</i>.
     * @throws FormProjectException if the label is not bound in this form.
    **/
    public abstract Form project(String label);

    public abstract Form projectDef(String label, Form defaultForm);

    public abstract boolean runtimeEquals(Form form);

    public abstract int runtimeHashCode();

    public abstract String runtimeToString();

    public abstract Service service();

    /**
     * Returns a string representation of this form.
     *
     * @return a string representation of this form.
    **/
    public abstract String toString();

    /**
     * Returns the contractual type of this form.
    **/
    public abstract ContractualType type();

    public abstract void push(VM vm);

}
